/*
 *  PROGRAM NAME:  multifario
 *
 *  (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
 *  CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.
 *
 *  Please refer to the LICENSE file in the top directory
 *
 *      author: Mike Henderson mhender@watson.ibm.com
 */

static char *id="@(#) $Id: ComputePlaneClip.c 249 2006-08-19 01:12:35Z mhender $";

#include <MFAtlas.h>

double XGreaterThanMinusOne(MFNVector u, MFErrorHandler e){return -MFNV_C(u,0,e)-1.;}
double XLessThanOne(MFNVector u, MFErrorHandler e){return MFNV_C(u,0,e)-1.;}
double YGreaterThanMinusOne(MFNVector u, MFErrorHandler e){return -MFNV_C(u,1,e)-1.;}
double YLessThanOne(MFNVector u, MFErrorHandler e){return MFNV_C(u,1,e)-1.;}

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  MFContinuationMethod H;
  MFErrorHandler e;

  e=MFCreateErrorHandler();

  M=MFIMFCreatePlane(e);
  n=MFIMF_N(M,e);
  Omega=MFNRegionCreateRectangle(-1.,-1.,1.,1.,e);

  u0=MFIMFVectorFactory(M,e);
  MFNVSetC(u0,0, 0.,e);
  MFNVSetC(u0,1, 0.,e);

  H=MFCreateHendersonsMethod(e);
  MFHendersonSetRealParameter(H,"epsilon",.1,e);
  MFHendersonSetIntegerParameter(H,"maxCharts",-1,e);
  MFHendersonSetIntegerParameter(H,"verbose",1,e);
  MFHendersonSetIntegerParameter(H,"page",1,e);
  MFHendersonSetIntegerParameter(H,"dumpToPlotFile",1,e);
  MFHendersonSetIntegerParameter(H,"dumpToCenterFile",0,e);
  MFHendersonSetFilename(H,"PlaneClip",e);

  MFHendersonAddClipF(H,XGreaterThanMinusOne,e);
  MFHendersonAddClipF(H,XLessThanOne,e);
  MFHendersonAddClipF(H,YGreaterThanMinusOne,e);
  MFHendersonAddClipF(H,YLessThanOne,e);

  S=MFComputeAtlas(H,M,Omega,u0,e);

  MFCloseAtlas(H,S,e);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(S,e);
  MFFreeImplicitMF(M,e);
  MFFreeNRegion(Omega,e);
  MFFreeNVector(u0,e);
  MFFreeContinuationMethod(H,e);

  MFFreeErrorHandler(e);

  return 0;
 }
